#!/usr/bin/perl

########################################################################
# metaf.pl -- CGI wrapper script to display METAR/TAF conversion results
#
# copyright (c) metaf2xml 2006-2007
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;

=head1 NAME

metaf.pl -- CGI wrapper script to display METAR/TAF conversion results

=head1 SYNOPSIS

as a standalone script:

  metaf.pl 'type=...&lang=...&format=...&source=...&mode=...&msg=...'

or with HTML:

  <form method="post" action="metaf.pl">
   <input name="type" ... >
   <input name="lang" ... >
   <input name="format" ... >
   <input name="source" ... >
   <input name="mode" ... >
   <input name="msg" ... >
  </form>

=head1 ARGUMENTS

The script uses the following 5 CGI parameters:

=over

=item B<type>

type of message to be parsed: metar (default), taf

=item B<lang>

language for output of script and display of parsed message: de (default), en,
es, ru

=item B<format>

output format: html (default), text, or xml (with style sheet reference)

=item B<source>

source for current weather: noaa (default), adds, addsds, file

=item B<mode>

which messages to fetch and how to display: latest/detailed (default),
all/summary

=item B<msg>

message to be parsed or the keyword C<FETCH> and a list of airport ids

=back

=head1 DESCRIPTION

This CGI script provides a web interface to display METAR/TAF conversion
results. It can be used with GET or POST methods. It uses the perl module
C<CGI>.

=cut

use CGI qw/:standard -no_xhtml/;

#
# restrict use of CGI:
#

# uploads are not allowed
$CGI::DISABLE_UPLOADS = 1;

=pod

The length of the body of the POST request (not the length of query string of
the URL) is restricted to 350 bytes.

=cut

$CGI::POST_MAX = 350;

=pod

The following directories and files are used by the script and can be adapted to
local needs by modifying the script (e.g. during install):

=over

=item B<$METAF2XML_BIN>

directory where the C<metaf2xml.pl> is expected (default: C</opt/metaf2xml/bin>)

=item B<$METAF2XML_XSL>

directory where the XSL scripts are located (default: C</opt/metaf2xml/share>)

=item B<$CGI_DATA_DIR>

directory for the local copies of messages (default: C</tmp>)

=item B<$CGI_TMP_DIR>

directory for the temporary XML file (default: C</tmp>)

=item B<$CURL>

path and options for C<curl> (default: C<curl -sSf --stderr ->)

=item B<$CURL_PROXY_OPTS>

options for C<curl> if a HTTP proxy should be used (default: empty)

=back

=cut

########################################################################
# config section, possibly changed by install
########################################################################
my $METAF2XML_BIN = '/opt/metaf2xml/bin';
my $METAF2XML_XSL = '/opt/metaf2xml/share';
my $CGI_DATA_DIR = '/tmp';
my $CGI_LOG_FILE = '/tmp/metaf2xml.log';
my $CGI_TMP_DIR = '/tmp';
my $CURL = 'curl -sSf --stderr -';
my $CURL_PROXY_OPTS = '';

#
# filter line(s) for $icao from $filename
#
sub process_file {
    my ($icao, $filename, $mode, $icao_wildchar) = @_;
    my ($found, $after_eof, $response, @arr);

    if (!open(FP, $filename)) {
        return ("$icao - ERROR - " . uc "$filename - $!");
    }

    $icao =~ s/ /|/g;
    $icao = $icao eq $icao_wildchar ? '[A-Z][A-Z0-9]{3}' : 'yudo'
        if $icao =~ /^ALL[MST]$/;

    # one could also set $/ instead of the "found" stuff but files with mixed
    # line endings would not work (should $/ be "\n\n" or "\r\n\r\n"?)
    $found = 0;
    $after_eof = 0;
    while (!$after_eof) {
        $_ = <FP> unless $after_eof;
        if (!defined $_) {
            $after_eof = 1;
            $_ = '';
        } else {
            s/[\n\r]+//g;
        }
        if (!$found) {
            next unless /^(?:METAR |TAF |SPECI )?(?:$icao) /;
            $found = 1;
            ($response = $_) =~ s/^(?:METAR |TAF |SPECI )//;
        } else {
            if (/^ *$/) {
                my $id;
                $response =~ s/ +/ /g;
                $id = ($response =~ /^(....)/)[0];
                # remove existing entries with same $id from @arr
                if ($mode eq 'summary') {
                    @arr = grep { $_ ne $response } @arr;
                } else {
                    @arr = grep { !/^$id / } @arr;
                }

                unshift @arr, $response;
                $found = 0;
            } else {
                chomp();
                $response .= ' ' . $_;
            }
        }
    }
    close FP;
    return @arr;
}

#
# fetch line(s) for $icao from NOAA/IWS
#
sub process_noaa_iws {
    my ($icao, $type, $mode) = @_;
    my ($response, $url);

    $icao =~ s/ /+/g;

    $url = 'http://weather.noaa.gov/cgi-bin/mget' . $type . '.pl?cccc=';

    if (!open(FP, "$CURL $CURL_PROXY_OPTS $url'$icao' || echo ERROR |")) {
        return ("$icao - ERROR - " . uc $!);
    }
    {
        local $/;
        $response = <FP>;
    }
    close FP;
    if ($response =~ /ERROR/) {
        $response =~ s/[\n\r]+/ /g;
        return ("$icao - ERROR - " . uc "$url - $response");
    }
    $response =~ s/[\n\r]+/ /g;
    $response =~ s/ +/ /g;
    return $response =~ /<font face=.couri[^>]*?> *(?:<pre>)? *(?:METAR|SPECI|TAF)? *([^<]+)/go;
}

#
# fetch line(s) for $icao from NOAA/ADDS
#
sub process_noaa_adds {
    my ($icao, $mode) = @_;
    my ($response, $url, $data);

    $icao =~ s/ /+/g;

    $url  = 'http://adds.aviationweather.gov/metars/';
    $data = 'std_trans=standard&chk_metars=on&hoursStr=' . ($mode eq 'summary' ? 'past 36 hours': 'most+recent+only') . '&chk_tafs=on&submitmet=Submit&station_ids=';

    if (!open(FP, "$CURL $CURL_PROXY_OPTS -d '$data$icao' $url || echo ERROR |"))
    {
        return ("$icao - ERROR - " . uc $!);
    }
    {
        local $/;
        $response = <FP>;
    }
    close FP;
    if ($response =~ /ERROR/) {
        $response =~ s/[\n\r]+/ /g;
        return ("$icao - ERROR - " . uc "$url - $response");
    }
    $response =~ s/[\n\r]+//g;
    $response =~ s/ +/ /g;
    $response =~ s/(<font face=.mono[^>]*?>)/$1METAR /igo;
    $response =~ s/(<pre>)/$1TAF /igo;
    return $response =~ /(?:<font face=.mono[^>]*?>|<pre>)([^<]+)/igo;
}

#
# fetch line(s) for $icao from NOAA/ADDS dataserver
#
sub process_noaa_addsds {
    my ($icao, $type, $mode) = @_;
    my ($response, $url);
    my (%issue_time, @raw_text, %metar, %taf);

    $icao =~ s/ /,/g;

    $url = 'http://weather.aero/dataserver1_1/httpparam?requestType=retrieve&format=xml&dataSource=';
    if ($type eq 'metar') {
        $url .= 'metars&fields=raw_text,station_id,observation_time,metar_type&hoursBeforeNow=' . ($mode eq 'summary' ? '72' : '4');
    } else {
        $url .= 'tafs&hoursBeforeNow=0&timeType=valid';
    }
    $url .= '&stationString=';

    if (!open(FP, "$CURL $CURL_PROXY_OPTS '$url$icao' || echo ERROR |")) {
        return ("$icao - ERROR - " . uc $!);
    }
    {
        local $/;
        $response = <FP>;
    }
    close FP;
    if ($response =~ /ERROR/) {
        $response =~ s/[\n\r]+/ /g;
        return ("$icao - ERROR - " . uc "$url - $response");
    }
    $response =~ s/[\n\r]+/ /g;
    $response =~ s/ +/ /g;
    if ($type eq 'metar') {
        # copy to hash, key is observation_time
        for ($response =~ /<((?:raw_text|station_id|observation_time|metar_type)>[^<]+)/igo)
        {
            /^([^>]+).(.*)/;
            if ($1 eq 'raw_text') {
                if (scalar keys %metar eq 4) {
                    my $key = $mode eq 'summary'
                              ? $metar{raw_text}
                              : "$metar{station_id},$metar{metar_type}";
                    if (   !exists $issue_time{$key}
                        || $issue_time{$key} lt $metar{observation_time})
                    {
                        $issue_time{$key} = $metar{observation_time};
                        $metar{raw_text} =~ s/^(?:METAR |SPECI )?/SPECI /
                            if $metar{metar_type} eq 'SPECI';

                        if ($metar{metar_type} eq 'METAR' || $mode eq 'summary')
                        {
                            push @raw_text, $metar{raw_text};
                        # not summary: add SPECI only if there is no METAR yet
                        } elsif (!exists $issue_time{"$metar{station_id},METAR"})
                        {
                            push @raw_text, $metar{raw_text};
                        }
                    }
                }
                %metar = ();
            }
            $metar{$1} = $2;
        }
        return @raw_text;
    }

    # copy to hash, key is issue_time
    for ($response =~ /<((?:raw_text|station_id|issue_time|valid_time_from|valid_time_to)>[^<]+)/igo)
    {
        /^([^>]+).(.*)/;
        if ($1 eq 'raw_text') {
            if (scalar keys %taf eq 5) {
                my $hours;
                $taf{valid_time_to} =~ s/.*T(..).*/$1/;
                $taf{valid_time_from} =~ s/.*T(..).*/$1/;
                $hours = $taf{valid_time_to} - $taf{valid_time_from};
                $hours += 24 if $hours < 0;
                my $key = "$taf{station_id},$hours";
                if (   !exists $issue_time{$key}
                    || $issue_time{$key} lt $taf{issue_time})
                {
                    $issue_time{$key} = $taf{issue_time};
                    push @raw_text, "$hours,$taf{raw_text}";
                }
            }
            %taf = ();
        }
        $taf{$1} = $2;
    }

    return @raw_text;
}

=pod

If there are any CGI errors the type of error is reported and the script
terminates.

=cut

if (cgi_error()) {
    print header({ status => cgi_error() }) .
        start_html({ title => cgi_error() }) . "\n" .
        b(cgi_error()) . "\n" .
        end_html();
    exit 0;
}

=pod

All parameters are checked to have one of the allowed values.

=cut

#
# determine type from request (default: metar)
#
my $type = param('type');
$type = 'metar' unless defined $type && $type eq 'taf';

=pod

If the parameter C<lang> is not set the first allowed value of the HTTP header
C<Accept-Language> is used.

=cut

#
# determine language from request, browser header (default: de)
#
my $lang = param('lang');
if (!defined $lang) {
    my $accept_lang = http('Accept-Language');
    if (defined $accept_lang) {
        for (split(/,/, $accept_lang)) {
            last if $_ =~ /^de/i;
            if ($_ =~ /^en/i) {
                $lang = 'en';
                last;
            }
            if ($_ =~ /^es/i) {
                $lang = 'es';
                last;
            }
            if ($_ =~ /^ru/i) {
                $lang = 'ru';
                last;
            }
        }
    }
}
$lang = 'de'
    unless defined $lang && ($lang eq 'en' || $lang eq 'es' || $lang eq 'ru');

#
# determine format from request (default: html)
#
my $format = param('format');
$format = 'html'
    unless defined $format && ($format eq 'text' || $format eq 'xml');

#
# determine source from request (default: noaa)
#
my $source = param('source');
if (-f "$CGI_DATA_DIR/metar.txt") {
    $source = 'file';
} else {
    $source = 'noaa'
        unless defined $source && ($source eq 'adds' || $source eq 'addsds');
}

#
# determine mode from request (default: latest)
#
my $mode = param('mode');
$mode = 'latest' unless defined $mode && $mode eq 'summary';

=pod

The parameter C<msg> is truncated to 500 characters. Letters are converted to
upper case. Any characters that are not allowed in a message are removed.

=cut

my $msg = param('msg') ? substr(param('msg'), 0, 500) : '';
$msg = uc $msg;
$msg =~ s/[^ -~]//gos; # avoid invalid XML and HTML

#
# determine if XML file name provided
#
my $xml_out_provided = param('xml_out_provided');
my $xml_out;
if (defined $xml_out_provided && ! -e "$CGI_TMP_DIR/metaf2xml-provided.xml") {
    $xml_out = "$CGI_TMP_DIR/metaf2xml-provided.xml";
} else {
    $xml_out_provided = undef;
    $xml_out = "$CGI_TMP_DIR/metaf2xml-$$.xml";
}

#
# log request
#
if (request_method() && open(LOG, '>>', $CGI_LOG_FILE)) {
    # anonymize IP address, we only want to group the requests
    my $addr = remote_addr();
    $addr =~ s/\.[0-9]+/.x/;
    $addr =~ s/[0-9]+$/x/;
    print LOG gmtime() . ' ' . $addr . ' ' . request_method()
              . " $type $lang $format $source $mode $msg\n";
    close LOG;
}

=pod

If C<msg> starts with C<METAR> or C<TAF> this message type is used instead of
the parameter C<type>.

=cut

#
# overwrite type if specified in message
#
$type = 'metar' if $msg =~ /^(?:METAR|SPECI) /;
$type = 'taf' if $msg =~ /^TAF /;
my $is_taf = $type eq 'taf' ? '-t ' : '';

#
# translated strings
#
my %trans = (
    title_en => 'METAR/TAF decoder',
    title_de => 'METAR/TAF-Übersetzer',
    title_es => 'METAR/TAF traductor',
    title_ru => 'МЕТАР/ТАФ декодер',
    header_en => 'METAR/TAF decoder',
    header_de => 'METAR/TAF-Übersetzer',
    header_es => 'METAR/TAF traductor',
    header_ru => 'МЕТАР/ТАФ декодер',
    infosrc_en => 'A good source for up-to-date METAR and TAF messages is',
    infosrc_de => 'Eine gute Quelle für aktuelle METAR- und TAF-Meldungen ist',
    infosrc_es => 'Una buena fuente por mensajes actuales de METAR y TAF es',
    infosrc_ru => 'Хороший источник для современных сообщений МЕТАР и ТАФ',
    infofetch_en => 'If the keyword "FETCH" and one or more Airport-Ids are used, messages are fetched from the selected source (or if this option is disabled from a file on the server where this web page comes from).',
    infofetch_de => 'Wird das Schlüsselwort "FETCH" und ein oder mehrere Flugplatz-Kürzel verwendet, werden Meldungen von der ausgewählten Quelle (oder, falls diese Option gesperrt ist, von einer Datei auf dem Server, von dem diese Web-Seite kommt) geholt.',
    infofetch_es => 'Si se usa la palabra clave "FETCH" y uno o mas ids de la estación, los mensajes serán traído de la fuente elegida (o si esta opción esta minusválida de un fichero del servidor de donde viene este página web).',
    infofetch_ru => 'Если ключевое слово "FETCH" и одни или более ид аэропорта используется, сообщения принесены из отобранного источника (или если этот выбор инвалиды от файла на сервере, откуда эта веб-страница прибывает).',
    example_en => 'examples for messages',
    example_de => 'Beispiele für Meldungen',
    example_es => 'ejemplos para mensajes',
    example_ru => 'примеры для сообщений',
    message_en => 'message',
    message_de => 'Meldung',
    message_es => 'mensaje',
    message_ru => 'сообщение',
    type_en  => 'is of type',
    type1_en => 'METAR', type2_en => 'TAF',
    type_de  => 'ist vom Typ',
    type1_de => 'METAR', type2_de => 'TAF',
    type_es  => 'es del typo',
    type1_es => 'METAR', type2_es => 'TAF',
    type_ru  => 'имеет тип',
    type1_ru => 'МЕТАР', type2_ru => 'ТАФ',
    source_en   => 'fetch from',
    source1_en  => 'NOAA/IWS', source2_en => 'NOAA/ADDS',
     source3_en => 'NOAA/ADDS dataserver',
    source_de   => 'hole von',
    source1_de  => 'NOAA/IWS', source2_de => 'NOAA/ADDS',
     source3_de => 'NOAA/ADDS dataserver',
    source_es   => 'trae de',
    source1_es  => 'NOAA/IWS', source2_es => 'NOAA/ADDS',
     source3_es => 'NOAA/ADDS dataserver',
    source_ru   => 'усилия от',
    source1_ru  => 'NOAA/IWS', source2_ru => 'NOAA/ADDS',
     source3_ru => 'NOAA/ADDS dataserver',
    language_en => 'language',
    language_de => 'Sprache',
    language_es => 'lengua',
    language_ru => 'язык',
    format_en   => 'output as',
    format1_en  => 'HTML', format2_en => 'text',
     format3_en => 'XML (+XSL -> HTML)',
    format_de   => 'gib aus als',
    format1_de  => 'HTML', format2_de => 'Text',
     format3_de => 'XML (+XSL -> HTML)',
    format_es   => 'presenta como',
    format1_es  => 'HTML', format2_es => 'texto',
     format3_es => 'XML (+XSL -> HTML)',
    format_ru   => 'показа в',
    format1_ru  => 'HTML', format2_ru => 'текст',
     format3_ru => 'XML (+XSL -> HTML)',
    mode_en     => 'fetch/display',
    mode1_en    => 'latest/detailed', mode2_en => 'all/summary',
    mode_de     => 'hole/zeige',
    mode1_de    => 'letzte/detailliert', mode2_de => 'alle/Übersicht',
    mode_es     => 'trae/presenta',
    mode1_es    => 'último/detallado', mode2_es => 'todos/resumen',
    mode_ru     => 'усилия/показа',
    mode1_ru    => 'послед./подробно', mode2_ru => 'все/обзор',
    decode_en   => 'decode',
    decode_de   => 'Übersetzen',
    decode_es   => 'traduce',
    decode_ru   => 'переводить',
    disclaimer_en => 'Warning: The information on this page may be out-dated, inaccurate, or both. It is not suited for use in aviation.',
    disclaimer_de => 'Warnung: Die Informationen auf dieser Seite könnten veraltet, falsch oder beides sein. Sie sind nicht für die Verwendung in der Luftfahrt geeignet.',
    disclaimer_es => 'Aviso: Las informaciones de esta página web puedan ser antiguas o falsas o los dos. No son apropiadas para el uso en la aviación.',
    disclaimer_ru => 'Предупреждение: Информация относительно этой страницы может быть устарелой, неточной, или оба. Этому не удовлетворяют для использования в авиации.',
);

#
# save message parameter
#
my $msg_orig = $msg;

=pod

If C<msg> starts with C<FETCH> and is followed by one or more airport ICAO codes
messages for the specified airports are fetched from the local files
C<$CGI_DATA_DIR/{metar,staf,taf}.txt> (if they exist) or the NOAA web site.

=cut

#
# process special contents of message
#
if ($msg =~ /^FETCH((?: [A-Z][A-Z0-9]{3})+)$/) {
    my ($icao, $response, @metar, @staf, @taf);

    $is_taf = '';
    $icao = $1;
    $icao =~ s/ //;

    if ($source ne 'file') {
        if ($source eq 'noaa') {
            @metar = process_noaa_iws $icao, 'metar', $mode;
            @staf  = process_noaa_iws $icao, 'staf', $mode;
            @taf   = process_noaa_iws $icao, 'taf', $mode;
        } elsif ($source eq 'adds') {
            for (process_noaa_adds $icao, $mode) {
                if (s/^METAR //) {
                    push @metar, $_;
                } elsif (s/^TAF //) {
                    push @taf, $_;
                } else {
                    push @metar, $_;
                }
            }
        } else {
            @metar = process_noaa_addsds $icao, 'metar', $mode;
            for (process_noaa_addsds $icao, 'taf', $mode) {
                if (s/^[1-9],//) {
                    push @staf, $_;
                } else {
                    s/^[0-9]+,//;
                    push @taf, $_;
                }
            }
        }
    } else {
        my ($metarfile, $staffile, $taffile, $tmp_icao);

        $metarfile = "$CGI_DATA_DIR/metar.txt";
        $staffile  = "$CGI_DATA_DIR/staf.txt";
        $taffile   = "$CGI_DATA_DIR/taf.txt";

        @metar = process_file $icao, $metarfile, $mode, 'ALLM';
        @staf  = process_file $icao, $staffile, $mode, 'ALLS';
        @taf   = process_file $icao, $taffile, $mode, 'ALLT';
        $icao = '.' if $icao =~ /^ALL[MST]$/;
    }

    $msg = '';
    for my $id (split / /, $icao) {
        $msg .= "\nMETAR\n" . join("\n", grep { /^(?:METAR |SPECI )?$id/ } @metar);
        $msg .= "\nTAF\n"   . join("\n", grep { /^$id/ } @staf);
        $msg .= "\nTAF\n"   . join("\n", grep { /^$id/ } @taf);
    }
    $msg =~ s/ +/ /g;
    $msg =~ s/ $//g;
    $msg =~ s/[^\n -~]/?/gos; # avoid invalid XML and HTML
    $msg =~ s/[\n]+/\n/g;
    $msg =~ s/^\n//;
}

if ($format eq 'xml') {
    my $xsl = $mode eq 'summary' ? 'sum-ui' : 'fullhtml';
    print "Content-Type: text/xml; charset=UTF-8\n\n";
    open(METAF2XML, "|$METAF2XML_BIN/metaf2xml.pl $is_taf -Xx - -S/metaf-$xsl.xsl -O'$type $lang $format $source $mode $msg_orig' 2>&1");
    print METAF2XML $msg;
    close METAF2XML;
    exit 0;
}

print header({charset => 'UTF-8'}) .
    start_html({ title => $trans{'title_'.$lang},
                 lang  => $lang,
                 dtd   => [ '-//W3C//DTD HTML 4.01 Transitional//EN',
                            'http://www.w3.org/TR/html4/loose.dtd' ],
                 meta  => { robots   => 'noindex,nofollow',
                            keywords =>
                     'metaf2xml,METAR,TAF,XML,aviation,weather,report,forecast',
                          },
                 head  => meta({ http_equiv => 'Content-Type',
                                 content    => 'text/html; charset=UTF-8'})
               }
              ) . "\n" .
    h2($trans{'header_'.$lang}) . "\n" .
    $trans{'infosrc_'.$lang} . ":\n" .
    kbd('http://weather.noaa.gov/weather/coded.html') . p . "\n" .
    $trans{'infofetch_'.$lang} . p . "\n" .
    start_form({ action => 'metaf.pl' }) .
    table({ border => 0, cellspacing => 0 }, "\n" .
        Tr(td({ colspan => 5 }, $trans{'example_'.$lang} . ':')) . "\n" .
        Tr("\n" .
           td('') . "\n" .
           td({ colspan => 4, style => 'font-family: monospace;' }, "\n" .
              'METAR YUDO 090600Z 00000KT CAVOK 22/15 Q1021' . "\n") . "\n"
        ) . "\n" .
        Tr("\n" .
           td('') . "\n" .
           td({ colspan => 4, style => 'font-family: monospace;' }, "\n" .
              'TAF YUDO 090600Z 090716 VRB03KT 8000 SKC PROB40 TEMPO 0708 0200 +TSRA FM0800 CAVOK' . "\n") . "\n"
        ) . "\n" .
        Tr("\n" .
           td('') . "\n" .
           td({ colspan => 4, style => 'font-family: monospace;' }, "\n" .
              'FETCH SBGL KJFK RJTT' . "\n") . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'message_'.$lang} . ':') . "\n" .
           td({ colspan => 4 }, "\n" .
              textfield({ name      => 'msg',
                          size      => 90,
                          maxlength => 500,
                          style     => 'font-family: monospace;',
                          value     => $msg_orig }) . "\n") . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'type_'.$lang} . ':') . "\n" .
           td(input({ type => 'radio', name => 'type', value => 'metar',
                      ($type eq 'metar' ? 'checked' : '') => undef }
                   ) .
              $trans{'type1_'.$lang}) . "\n" .
           td({ colspan => 3 },
              input({ type => 'radio', name => 'type', value => 'taf',
                      ($type eq 'taf' ? 'checked' : '') => undef }
                   ) .
              $trans{'type2_'.$lang}) . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'source_'.$lang} . ':') . "\n" .
           td(input({ type => 'radio', name => 'source', value => 'noaa',
                      ($source eq 'noaa' ? 'checked' : '') => undef,
                      ($source eq 'file' ? 'readonly' : '') => undef,
                      ($source eq 'file' ? 'disabled' : '') => undef,
                    }
                   ) .
              $trans{'source1_'.$lang}) . "\n" .
           td(input({ type => 'radio', name => 'source', value => 'adds',
                      ($source eq 'adds' ? 'checked' : '') => undef,
                      ($source eq 'file' ? 'readonly' : '') => undef,
                      ($source eq 'file' ? 'disabled' : '') => undef,
                    }
                   ) .
              $trans{'source2_'.$lang}) . "\n" .
           td({ colspan => 2 },
              input({ type => 'radio', name => 'source', value => 'addsds',
                      ($source eq 'addsds' ? 'checked' : '') => undef,
                      ($source eq 'file' ? 'readonly' : '') => undef,
                      ($source eq 'file' ? 'disabled' : '') => undef,
                    }
                   ) .
              $trans{'source3_'.$lang}) . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'language_'.$lang} . ':') . "\n" .
           td({ width => '16%' },
              input({ type => 'radio', name => 'lang', value => 'de',
                      ($lang eq 'de' ? 'checked' : '') => undef }
                   ) .
              'Deutsch') . "\n" .
           td({ width => '19%' },
              input({ type => 'radio', name => 'lang', value => 'en',
                      ($lang eq 'en' ? 'checked' : '') => undef }
                   ) .
              'English') . "\n" .
           td({ width => '25%' },
              input({ type => 'radio', name => 'lang', value => 'es',
                      ($lang eq 'es' ? 'checked' : '') => undef }
                   ) .
              'Español') . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'ru',
                      ($lang eq 'ru' ? 'checked' : '') => undef }
                   ) .
              'Русский') . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'format_'.$lang} . ':') . "\n" .
           td(input({ type => 'radio', name => 'format', value => 'html',
                      ($format eq 'html' ? 'checked' : '') => undef }
                   ) .
              $trans{'format1_'.$lang}) . "\n" .
           td(input({ type => 'radio', name => 'format', value => 'text',
                      ($format eq 'text' ? 'checked' : '') => undef }
                   ) .
              $trans{'format2_'.$lang}) . "\n" .
           td({ colspan => 2 },
              input({ type => 'radio', name => 'format', value => 'xml',
                      ($format eq 'xml' ? 'checked' : '') => undef }
                   ) .
              $trans{'format3_'.$lang}) . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'mode_'.$lang} . ':') . "\n" .
           td(input({ type => 'radio', name => 'mode', value => 'latest',
                      ($mode eq 'latest' ? 'checked' : '') => undef }
                   ) .
              $trans{'mode1_'.$lang}) . "\n" .
           td({ colspan => 3 },
              input({ type => 'radio', name => 'mode', value => 'summary',
                      ($mode eq 'summary' ? 'checked' : '') => undef }
                   ) .
              $trans{'mode2_'.$lang}) . "\n"
        ) . "\n"
    ) . "\n" .
    submit( value => $trans{'decode_'.$lang }) .
    end_form . p . "\n";

if ($msg) {
    my $response;
    my $sum = $mode eq 'summary' ? 'sum-' : '';

    print b($trans{'disclaimer_'.$lang}), p, "\n";

    open(METAF2XML, "|$METAF2XML_BIN/metaf2xml.pl $is_taf -Xx $xml_out 2>&1");
    print METAF2XML $msg;
    close METAF2XML;
    open(XSLTPROC, "xsltproc --nonet --nowrite --nomkdir --stringparam lang $lang $METAF2XML_XSL/metaf-$sum$format.xsl $xml_out 2>&1 |");
    {
        local $/;
        $response = <XSLTPROC>;
    }
    close XSLTPROC;
    unlink $xml_out unless defined $xml_out_provided;
    if ($format eq 'html') {
        print $response;
    } else {
        print table({ border => 1 }, "\n" .
              Tr("\n" .
                 td("\n" .
                    pre({ style => 'margin: 0;' }, $response)) . "\n") . "\n");
    }
}

print comment('metaf.pl: $Id: metaf.pl,v 1.42 2007/03/04 17:21:53 metaf2xml Exp $'), "\n", p,
      "copyright (c) 2006-2007 metaf2xml @\n",
      a({ href => 'http://metaf2xml.sourceforge.net/' }, "\n" .
      img({ src    =>
              'http://sflogo.sourceforge.net/sflogo.php?group_id=168043&type=1',
            border => 0,
            alt    => 'SourceForge.net Logo' }) . "\n") . "\n" .
      end_html();
